import argparse
import numpy as np
import subprocess
from pathlib import Path
import torch
from adabmDCA.fasta import get_tokens, write_fasta
from adabmDCA.parser import add_args_train, add_args_reintegration
from adabmDCA.utils import get_device, get_dtype
from adabmDCA.dataset import DatasetDCA

def create_parser():
    parser = argparse.ArgumentParser(description='Reintegrate a DCA model.')
    parser = add_args_train(parser)
    parser = add_args_reintegration(parser)
    return parser

def main():    
    parser = create_parser()
    args = parser.parse_args()
    
    print("\n" + "".join(["*"] * 10) + f" Training a reintegrated DCA model " + "".join(["*"] * 10) + "\n")
    print("Generating the reintegrated dataset...")
    
    device = get_device(args.device, message=False)
    dtype = get_dtype(args.dtype)
    tokens = get_tokens(args.alphabet)
    
    # Create the folder where to save the model
    folder = Path(args.output)
    folder.mkdir(parents=True, exist_ok=True)
    
    dataset_nat = DatasetDCA(
        path_data=args.data,
        path_weights=args.weights,
        alphabet=args.alphabet,
        clustering_th=args.clustering_seqid,
        no_reweighting=args.no_reweighting,
        device=device,
        dtype=dtype,
        message=False,
    )
    
    dataset_reint = DatasetDCA(
        path_data=args.reint,
        alphabet=args.alphabet,
        no_reweighting=True,
        device=device,
        dtype=dtype,
        message=False,
    )
    
    # concatenate the two datasets
    msa = torch.cat((dataset_nat.data, dataset_reint.data), dim=0)
    msa_names = np.append(dataset_nat.names, dataset_reint.names)
    k = (args.lambda_ * dataset_nat.get_effective_size()) / len(dataset_reint)
    # Import the adjustment vector
    with open(args.adj, "r") as f:
        adjust = torch.tensor([float(x) for x in f.read().split()], device=device, dtype=dtype)
    # Conncatenate the weights
    weights = torch.cat((dataset_nat.weights.view(-1), k * adjust), dim=0).unsqueeze(1)
    
    # Save the new dataset
    args.label = f"{args.label}-lambda_{args.lambda_}" if args.label is not None else f"lambda_{args.lambda_}"
    path_msa = folder / Path(f"{args.label}_msa.fasta")
    write_fasta(
        fname=path_msa,
        headers=msa_names,
        sequences=msa.cpu().numpy(),
        numeric_input=True,
        remove_gaps=False,
        tokens=tokens,
    )
    path_weights = folder / Path(f"{args.label}_weights.dat")
    np.savetxt(path_weights, weights.cpu().numpy())
    
    # launch the training
    train_command = [
        "adabmDCA",
        "train",
        "--data", str(path_msa),
        "--weights", str(path_weights),
        "--output", str(folder),
        "--device", str(args.device),
        "--dtype", str(args.dtype),
        "--alphabet", str(args.alphabet),
        "--model", str(args.model),
        "--clustering_seqid", str(args.clustering_seqid),
        "--lr", str(args.lr),
        "--label", str(args.label),
        "--nsweeps", str(args.nsweeps),
        "--nchains", str(args.nchains),
        "--target", str(args.target),
        "--nepochs", str(args.nepochs),
        "--pseudocount", str(0),
        "--seed", str(args.seed),
        "--checkpoints", str(args.checkpoints),
        "--target_acc_rate", str(args.target_acc_rate),
        "--gsteps", str(args.gsteps),
        "--factivate", str(args.factivate),
        "--density", str(args.density),
        "--drate", str(args.drate),
    ]
    if args.no_reweighting:
        train_command.append("--no_reweighting")
    if args.wandb:
        train_command.append("--wandb")
    if args.path_params is not None:
        train_command.append("--path_params")
        train_command.append(str(args.path_params))
    if args.path_chains is not None:
        train_command.append("--path_chains")
        train_command.append(str(args.path_chains))
    
    subprocess.run(train_command, check=True)
    
if __name__ == "__main__":
    main()
    


import torch


@torch.jit.script
def _one_hot(x: torch.Tensor, num_classes: int = -1, dtype: torch.dtype = torch.float32):
   
    if x.dim() != 2:
        raise ValueError("Input tensor x must be 2D")
    
    if num_classes < 0:
        num_classes = int(x.max() + 1)
    res = torch.zeros(x.shape[0], x.shape[1], num_classes, device=x.device, dtype=dtype)
    tmp = torch.meshgrid(
        torch.arange(x.shape[0], device=x.device),
        torch.arange(x.shape[1], device=x.device),
        indexing="ij",
    )
    index = (tmp[0], tmp[1], x)
    values = torch.ones(x.shape[0], x.shape[1], device=x.device, dtype=dtype)
    res.index_put_(index, values)
    
    return res


def one_hot(x: torch.Tensor, num_classes: int = -1, dtype: torch.dtype = torch.float32):
    """A fast one-hot encoding function faster than the PyTorch one working with torch.int32 and returning a float Tensor.
    Works only for 2D tensors.
    
    Args:
        x (torch.Tensor): Input tensor to be one-hot encoded.
        num_classes (int, optional): Number of classes. If -1, the number of classes is inferred from the input tensor. Defaults to -1.
        dtype (torch.dtype, optional): Data type of the output tensor. Defaults to torch.float32.
        
    Returns:
        torch.Tensor: One-hot encoded tensor.
    """
    return _one_hot(x, num_classes, dtype)


@torch.jit.script
def multinomial_one_hot(logits: torch.Tensor) -> torch.Tensor:
    """Sample from a multinomial distribution and return one-hot encoded samples."""
    assert logits.dim() == 2, "Input logits must be a 2D tensor (N, q) where N is the number of samples and q is the number of classes."
    batch_size = logits.size(0)
    one_hot_samples = torch.zeros_like(logits)
    probs = torch.softmax(logits, dim=-1)
    cdf = torch.cumsum(probs, dim=1)
    u = torch.rand(batch_size, 1, device=logits.device, dtype=logits.dtype) * cdf[:, -1].unsqueeze(1)
    indices = torch.searchsorted(cdf, u, right=True)
    one_hot_samples.scatter_(-1, indices, 1.0)
    
    return one_hot_samples


from typing import Dict
import itertools

import torch

from adabmDCA.functional import one_hot


@torch.jit.script
def _compute_energy_sequence(
    x: torch.Tensor,
    params: Dict[str, torch.Tensor],
) -> torch.Tensor:
    L, q = params["bias"].shape
    x_oh = x.ravel()
    bias_oh = params["bias"].ravel()
    couplings_oh = params["coupling_matrix"].view(L * q, L * q)
    energy = - x_oh @ bias_oh - 0.5 * x_oh @ (couplings_oh @ x_oh)
    
    return energy


def compute_energy(
    X: torch.Tensor,
    params: Dict[str, torch.Tensor],
) -> torch.Tensor:
    """Compute the DCA energy of the sequences in X.
    
    Args:
        X (torch.Tensor): Sequences in one-hot encoding format.
        params (Dict[str, torch.Tensor]): Parameters of the model.
    
    Returns:
        torch.Tensor: DCA Energy of the sequences.
    """
    
    if X.dim() != 3:
        raise ValueError("Input tensor X must be 3-dimensional of size (_, L, q)")
    
    return torch.vmap(_compute_energy_sequence, in_dims=(0, None))(X, params)


def _update_weights_AIS(
    prev_params: Dict[str, torch.Tensor],
    curr_params: Dict[str, torch.Tensor],
    chains: torch.Tensor,
    log_weights: torch.Tensor,
) -> torch.Tensor:
    """Update the weights used during  the trajectory Annealed Importance Sampling (AIS) algorithm.

    Args:
        prev_params (Dict[str, torch.Tensor]): Params at time t-1.
        curr_params (Dict[str, torch.Tensor]): Params at time t.
        chains (torch.Tensor): Chains at time t-1.
        log_weights (torch.Tensor): Log-weights at time t-1.

    Returns:
        torch.Tensor: Log-weights at time t.
    """
    energy_prev = compute_energy(chains, prev_params)
    energy_curr = compute_energy(chains, curr_params)
    log_weights += energy_prev - energy_curr
    
    return log_weights


def _compute_ess(log_weights: torch.Tensor) -> float:
    """Computes the Effective Sample Size of the chains.

    Args:
        log_weights: log-weights of the chains.
    """
    lwc = log_weights - log_weights.min()
    numerator = torch.square(torch.mean(torch.exp(-lwc))).item()
    denominator = torch.mean(torch.exp(-2.0 * lwc)).item()

    return numerator / denominator


@torch.jit.script
def _compute_log_likelihood(
    fi: torch.Tensor,
    fij: torch.Tensor,
    params: Dict[str, torch.Tensor],
    logZ: float,
) -> float:
    
    mean_energy_data = - torch.sum(fi * params["bias"]) - 0.5 * torch.sum(fij * params["coupling_matrix"])

    return - mean_energy_data.item() - logZ


def compute_log_likelihood(
    fi: torch.Tensor,
    fij: torch.Tensor,
    params: Dict[str, torch.Tensor],
    logZ: float,
) -> float:
    """Compute the log-likelihood of the model.

    Args:
        fi (torch.Tensor): Single-site frequencies of the data.
        fij (torch.Tensor): Two-site frequencies of the data.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        logZ (float): Log-partition function of the model.

    Returns:
        float: Log-likelihood of the model.
    """
    return _compute_log_likelihood(fi, fij, params, logZ)


def enumerate_states(
    L: int,
    q: int,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Enumerate all possible states of a system of L sites and q states.

    Args:
        L (int): Number of sites.
        q (int): Number of states.
        device (torch.device, optional): Device to store the states. Defaults to "cpu".

    Returns:
        torch.Tensor: All possible states.
    """
    if q**L > 5**11:
        raise ValueError("The number of states is too large to enumerate.")
    
    all_states = torch.tensor(list(itertools.product(range(q), repeat=L)), device=device)
    return one_hot(all_states, q)


def compute_logZ_exact(
    all_states: torch.Tensor,
    params: Dict[str, torch.Tensor],
) -> float:
    """Compute the log-partition function of the model.

    Args:
        all_states (torch.Tensor): All possible states of the system.
        params (Dict[str, torch.Tensor]): Parameters of the model.

    Returns:
        float: Log-partition function of the model.
    """
    energies = compute_energy(all_states, params)
    logZ = torch.logsumexp(-energies, dim=0)
    
    return logZ.item()


def compute_entropy(
    chains: torch.Tensor,
    params: Dict[str, torch.Tensor],
    logZ: float,
) -> float:
    """Compute the entropy of the DCA model.

    Args:
        chains (torch.Tensor): Chains that are supposed to be an equilibrium realization of the model.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        logZ (float): Log-partition function of the model.

    Returns:
        float: Entropy of the model.
    """
    mean_energy = compute_energy(chains, params).mean()
    entropy = mean_energy + logZ
    
    return entropy.item()


def _get_acceptance_rate(
    prev_params: Dict[str, torch.Tensor],
    curr_params: Dict[str, torch.Tensor],
    prev_chains: torch.Tensor,
    curr_chains: torch.Tensor,
) -> float:
    """Compute the acceptance rate of swapping the configurations between two models alonge the training.

    Args:
        prev_params (Dict[str, torch.Tensor]): Parameters at time t-1.
        curr_params (Dict[str, torch.Tensor]): Parameters at time t.
        prev_chains (torch.Tensor): Chains at time t-1.
        curr_chains (torch.Tensor): Chains at time t.

    Returns:
        float: Acceptance rate of swapping the configurations between two models alonge the training.
    """
    nchains = len(prev_chains)
    delta_energy = (
        - compute_energy(curr_chains, prev_params)
        + compute_energy(prev_chains, prev_params)
        + compute_energy(curr_chains, curr_params)
        - compute_energy(prev_chains, curr_params)
    )
    swap = torch.exp(delta_energy) > torch.rand(size=(nchains,), device=delta_energy.device)
    acceptance_rate = swap.float().mean().item()
    
    return acceptance_rate


@torch.jit.script
def _tap_residue(
    idx: int,
    mag: torch.Tensor,
    params: Dict[str, torch.Tensor],
) -> torch.Tensor:
    N, L, q = mag.shape
    coupling_residue = params["coupling_matrix"][idx] # (q, L, q)
    bias_residue = params["bias"][idx] # (q,)
    mag_i = mag[:, idx] # (n, q)
    
    mf_term = bias_residue + mag.view(N, L * q) @ coupling_residue.view(q, L * q).T
    reaction_term_temp = (
        0.5 * coupling_residue.view(1, q, L, q) + # (1, q, L, q)
        (torch.tensordot(mag_i, coupling_residue, dims=[[1], [0]]) * mag).sum(dim=2).view(N, 1, L, 1) - # nd,djc,njc->nj
        0.5 * torch.einsum("njc,ajc->naj", mag, coupling_residue).view(N, q, L, 1) -                    # njc,ajc->naj
        torch.tensordot(mag_i, coupling_residue, dims=[[1], [0]]).view(N, 1, L, q)                      # nd,djb->njb
    )
    reaction_term = (
        (reaction_term_temp * coupling_residue.view(1, q, L, q)) * mag.view(N, 1, L, q)
    ).sum(dim=3).sum(dim=2) # najb,ajb,njb->na
    tap_residue = torch.softmax(mf_term + reaction_term, dim=1)
    
    return tap_residue


def _sweep_tap(
    residue_idxs: torch.Tensor,
    mag: torch.Tensor,
    params: Dict[str, torch.Tensor],    
) -> torch.Tensor:
    """Updates the magnetizations using the TAP equations.

    Args:
        residue_idxs (torch.Tensor): List of residue indices in random order.
        mag (torch.Tensor): Magnetizations of the residues.
        params (Dict[str, torch.Tensor]): Parameters of the model.

    Returns:
        torch.Tensor: Updated magnetizations.
    """
    for idx in residue_idxs:
        mag[:, idx] = _tap_residue(idx, mag, params)  
    
    return mag


def iterate_tap(
    mag: torch.Tensor,
    params: Dict[str, torch.Tensor],
    max_iter: int = 500,
    epsilon: float = 1e-4,
):
    """Iterates the TAP equations until convergence.

    Args:
        mag (torch.Tensor): Initial magnetizations.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        max_iter (int, optional): Maximum number of iterations. Defaults to 2000.
        epsilon (float, optional): Convergence threshold. Defaults to 1e-6.

    Returns:
        torch.Tensor: Fixed point magnetizations of the TAP equations.
    """
    # ensure that mag is a 3D tensor
    if mag.dim() != 3:
        raise ValueError("Input tensor mag must be 3-dimensional of size (_, L, q)")
    
    mag_ = mag.clone()
    iterations = 0
    while True:
        mag_old = mag_.clone()
        mag_ = _sweep_tap(torch.randperm(mag_.shape[1]), mag_, params)
        diff = torch.abs(mag_old - mag_).max()
        iterations += 1
        if diff < epsilon or iterations > max_iter:
            break
    
    return mag_


import argparse
from pathlib import Path

import torch

from adabmDCA.fasta import get_tokens
from adabmDCA.utils import get_device, get_dtype
from adabmDCA.io import load_params, import_from_fasta
from adabmDCA.fasta import encode_sequence
from adabmDCA.functional import one_hot
from adabmDCA.statmech import compute_energy
from adabmDCA.parser import add_args_energies


# import command-line input arguments
def create_parser():
    parser = argparse.ArgumentParser(description="Computes the DCA energies of a sequence dataset.")
    parser = add_args_energies(parser)
    
    return parser


def main():    
    # Parse arguments
    parser = create_parser()
    args = parser.parse_args()
    
    print("\n" + "".join(["*"] * 10) + f" Computing DCA energies " + "".join(["*"] * 10) + "\n")
    
    # Set the device
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)
    
    # Check if the data file exists
    if not Path(args.data).exists():
        raise FileNotFoundError(f"Data file {args.data} not found.")
    
    # Check if the parameters file exists
    if not Path(args.path_params).exists():
        raise FileNotFoundError(f"Parameters file {args.path_params} not found.")
    
    # import data
    tokens = get_tokens(args.alphabet)
    names, sequences = import_from_fasta(args.data)
    data = encode_sequence(sequences, tokens)
    data = torch.tensor(data, device=device, dtype=torch.int32)
    
    # import parameters and compute DCA energies
    print(f"Loading parameters from {args.path_params}...")
    params = load_params(args.path_params, tokens=tokens, device=device, dtype=dtype)
    q = params["bias"].shape[1]
    data = one_hot(data, num_classes=q).to(dtype)
    print(f"Computing DCA energies...")
    energies = compute_energy(data, params).cpu().numpy()
    
    # Save results in a file
    print("Saving results...")
    folder = args.output
    folder.mkdir(parents=True, exist_ok=True)
    fname_out = folder / Path(args.data.stem + "_energies.fasta")
    with open(fname_out, "w") as f:
        for n, s, e in zip(names, sequences, energies):
            f.write(f">{n} | DCAenergy: {e:.3f}\n")
            f.write(f"{s}\n")
    
    print(f"Process completed. Output saved in {fname_out}")
    
    
if __name__ == "__main__":
    main()


import torch
from tqdm.autonotebook import trange

@torch.jit.script
def has_neighbours(seq: torch.Tensor, db: torch.Tensor, threshold: float) -> bool:
    """
    Check if a sequence 'seq' has neighbours in the database 'db', that is, sequences
    that have pairwise sequence identity greater than 'threshold'.
    """
    if len(db) == 0:
        return False
    else:
        seqid = (seq == db).float().sum(dim=1) / seq.shape[0]
        return bool((seqid > threshold).any().item())


def split_train_test(
    headers: list[str],
    X: torch.Tensor,
    seqid_th: float,
    rnd_gen: torch.Generator | None = None,
) -> tuple[list, torch.Tensor, list, torch.Tensor]:
    """Splits X into two sets, T and S, such that no sequence in S has more than
    'seqid_th' fraction of its residues identical to any sequence in T.
    
    Args:
        headers (list[str]): List of sequence headers.
        X (torch.Tensor): Encoded input MSA.
        seqid_th (float): Threshold sequence identity.
        rnd_gen (torch.Generator, optional): Random number generator. Defaults to None.

    Returns:
        tuple[list, torch.Tensor, list, torch.Tensor]: Training and test sets.
    """
    T_mask = torch.tensor([False] * len(X), dtype=torch.bool, device=X.device)
    S_mask = torch.tensor([False] * len(X), dtype=torch.bool, device=X.device)
    perm = torch.randperm(len(X), generator=rnd_gen, device=X.device)
    X = X[perm]
    for i in trange(len(X), desc="Train/test splitting", leave=False):
        has_neighbours_in_S = has_neighbours(X[i], X[S_mask], seqid_th)
        has_neighbours_in_T = has_neighbours(X[i], X[T_mask], seqid_th)
        if torch.rand(1, generator=rnd_gen, device=X.device).item() < 0.5:
            if not has_neighbours_in_T:
                S_mask[i] = True
            elif not has_neighbours_in_S:
                T_mask[i] = True
        else:
            if not has_neighbours_in_S:
                T_mask[i] = True
            elif not has_neighbours_in_T:
                S_mask[i] = True
    
    T_mask = T_mask.cpu().numpy()
    S_mask = S_mask.cpu().numpy()
    T_headers = headers[T_mask]
    T = X[T_mask]
    S_headers = headers[S_mask]
    S = X[S_mask]
    
    if len(T) > len(S):
        return T_headers, T, S_headers, S
    else:
        return S_headers, S, T_headers, T
    
    
def prune_redundant_sequences(
    headers: list[str],
    X: torch.Tensor,
    seqid_th: float,
    rnd_gen: torch.Generator | None = None,
) -> tuple[list, torch.Tensor]:
    """Prunes sequences from X such that no sequence has more than 'seqid_th' fraction of its residues identical to any other sequence in the set.

    Args:
        headers (list[str]): List of sequence headers.
        X (torch.Tensor): Encoded input MSA.
        seqid_th (float): Threshold sequence identity.
        rnd_gen (torch.Generator, optional): Random generator. Defaults to None.

    Returns:
        tuple[list, torch.Tensor]: Pruned sequences.
    """
    perm = torch.randperm(len(X), generator=rnd_gen, device=X.device)
    X = X[perm]
    U_mask = torch.tensor([False] * len(X), dtype=torch.bool, device=X.device)
    for i in trange(len(X), desc="Pruning redundant sequences", leave=False):
        has_neighbours_in_U = has_neighbours(X[i], X[U_mask], seqid_th)
        if not has_neighbours_in_U:
            U_mask[i] = True
    U_mask = U_mask.cpu().numpy()
    
    return headers[U_mask], X[U_mask]


def run_cobalt(
    headers: list[str],
    X: torch.Tensor,
    t1: float,
    t2: float,
    t3: float,
    max_train: int | None,
    max_test: int | None,
    rnd_gen: torch.Generator | None = None,
) -> tuple[list, torch.Tensor, list, torch.Tensor]:
    """
    Runs the Cobalt algorithm to split the input MSA into training and test sets.
    
    Args:
        headers (list[str]): List of sequence headers.
        X (torch.Tensor): Encoded input MSA.
        t1 (float): No sequence in S has more than this fraction of its residues identical to any sequence in T.
        t2 (float): No pair of test sequences has more than this value fractional identity.
        t3 (float): No pair of training sequences has more than this value fractional identity.
        max_train (int | None): Maximum number of sequences in the training set.
        max_test (int | None): Maximum number of sequences in the test set.
        rnd_gen (torch.Generator, optional): Random number generator. Defaults to None.

    Returns:
        tuple[list, torch.Tensor, list, torch.Tensor]: Training and test sets.
    """
    # Cobalt step 1
    headers_train, train, headers_test, test = split_train_test(headers, X, t1, rnd_gen)
    # Cobalt step 2
    if len(test) > 0:
        headers_test, test = prune_redundant_sequences(headers_test, test, t2, rnd_gen)
    # Cobalt step 3
    if len(train) > 0:
        headers_train, train = prune_redundant_sequences(headers_train, train, t3, rnd_gen)
    
    if max_train is not None and len(train) > max_train:
        train_ids = torch.randperm(len(train), generator=rnd_gen)[:max_train]
        headers_train = headers_train[train_ids]
        train = train[train_ids]
    if max_test is not None and len(test) > max_test:
        test_ids = torch.randperm(len(test), generator=rnd_gen)[:max_test]
        headers_test = headers_test[test_ids]
        test = test[test_ids]
        
    return headers_train, train, headers_test, test


from typing import Dict, Callable

import torch
from torch.nn.functional import one_hot
from adabmDCA.functional import multinomial_one_hot


def gibbs_mutate(
    chains: torch.Tensor,
    num_mut: int,
    params: Dict[str, torch.Tensor],
    beta: float,
) -> torch.Tensor:
    """Attempts to perform num_mut mutations using the Gibbs sampler.

    Args:
        chains (torch.Tensor): One-hot encoded sequences.
        num_mut (int): Number of proposed mutations.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        beta (float): Inverse temperature.

    Returns:
        torch.Tensor: Updated chains.
    """
    N, L, q = chains.shape
    idx_array = torch.randint(0, L, (num_mut,), device=chains.device)
    for i in idx_array:
        # Select the couplings attached to the residue we are considering (i) and flatten along the other residues ({j})
        couplings_residue = params["coupling_matrix"][i].view(q, L * q)
        # Update the chains
        logit_residue = beta * (params["bias"][i].unsqueeze(0) + chains.reshape(N, L * q) @ couplings_residue.T) # (N, q)
        chains[:, i, :] = multinomial_one_hot(logit_residue)
        
    return chains


def gibbs_sampling(
    chains: torch.Tensor,
    params: Dict[str, torch.Tensor],
    nsweeps: int,
    beta: float = 1.0,
) -> torch.Tensor:
    """Gibbs sampling.
    
    Args:
        chains (torch.Tensor): Initial chains.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        nsweeps (int): Number of sweeps, where one sweep corresponds to attempting L mutations.
        beta (float, optional): Inverse temperature. Defaults to 1.0.
    
    Returns:
        torch.Tensor: Updated chains.
    """
    L = params["bias"].shape[0]
    chains_mutate = chains.clone() # avoids to modify the chains inplace
    for _ in torch.arange(nsweeps):
        chains_mutate = gibbs_mutate(chains_mutate, L, params, beta)

    return chains_mutate


def _get_deltaE(
        idx: int | torch.Tensor,
        chain: torch.Tensor,
        residue_old: torch.Tensor,
        residue_new: torch.Tensor,
        params: Dict[str, torch.Tensor],
        L: int,
        q: int,
    ) -> torch.Tensor:
    """Computes the energy difference between the new and old residue at position i.
    
    Args:
        idx (int | torch.Tensor): Index of the residue to mutate.
        chain (torch.Tensor): Current chain.
        residue_old (torch.Tensor): One-hot encoded representation of the old residue.
        residue_new (torch.Tensor): One-hot encoded representation of the new residue.
        params (Dict[str, torch.Tensor]): Model parameters.
        L (int): Length of the sequence.
        q (int): Number of possible states.
    
    Returns:
        torch.Tensor: Energy difference between the new and old residue.
    """
    coupling_residue = chain.view(-1, L * q) @ params["coupling_matrix"][:, :, idx, :].view(L * q, q) # (N, q)
    E_old = - residue_old @ params["bias"][idx] - torch.vmap(torch.dot, in_dims=(0, 0))(coupling_residue, residue_old)
    E_new = - residue_new @ params["bias"][idx] - torch.vmap(torch.dot, in_dims=(0, 0))(coupling_residue, residue_new)
    
    return E_new - E_old


def metropolis_mutate(
    chains: torch.Tensor,
    num_mut: int,
    params: Dict[str, torch.Tensor],
    beta: float,
) -> torch.Tensor:
    """Attempts to perform num_mut mutations using the Metropolis sampler.

    Args:
        chains (torch.Tensor): One-hot encoded sequences.
        num_mut (int): Number of proposed mutations.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        beta (float): Inverse temperature.

    Returns:
        torch.Tensor: Updated chains.
    """
    N, L, q = chains.shape
    idx_array = torch.randint(0, L, (num_mut,), device=chains.device)
    for i in idx_array:
        res_old = chains[:, i, :]
        res_new = one_hot(torch.randint(0, q, (N,), device=chains.device), num_classes=q).type(chains.dtype)
        delta_E = _get_deltaE(i, chains, res_old, res_new, params, L, q)
        accept_prob = torch.exp(- beta * delta_E).unsqueeze(-1)
        chains[:, i, :] = torch.where(accept_prob > torch.rand((N, 1), device=chains.device, dtype=chains.dtype), res_new, res_old)

    return chains
    

def metropolis(
    chains: torch.Tensor,
    params: Dict[str, torch.Tensor],
    nsweeps: int,
    beta: float = 1.0,
) -> torch.Tensor:
    """Metropolis sampling.

    Args:
        chains (torch.Tensor): One-hot encoded sequences.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        nsweeps (int): Number of sweeps to be performed, where one sweep corresponds to attempting L mutations.
        beta (float, optional): Inverse temperature. Defaults to 1.0.

    Returns:
        torch.Tensor: Updated chains.
    """
    L = params["bias"].shape[0]
    chains_mutate = chains.clone() # avoids to modify the chains inplace
    for _ in range(nsweeps):
        chains_mutate = metropolis_mutate(chains_mutate, L, params, beta)

    return chains_mutate


def get_sampler(sampling_method: str) -> Callable:
    """Returns the sampling function corresponding to the chosen method.

    Args:
        sampling_method (str): String indicating the sampling method. Choose between 'metropolis' and 'gibbs'.

    Raises:
        KeyError: Unknown sampling method.

    Returns:
        Callable: Sampling function.
    """
    if sampling_method == "gibbs":
        return gibbs_sampling
    elif sampling_method == "metropolis":
        return metropolis
    else:
        raise KeyError("Unknown sampling method. Choose between 'metropolis' and 'gibbs'.")



from pathlib import Path
from typing import Callable, Dict
import time

import torch

from adabmDCA.stats import get_correlation_two_points
from adabmDCA.training import train_graph
from adabmDCA.utils import get_mask_save
from adabmDCA.stats import get_freq_single_point, get_freq_two_points, get_correlation_two_points
from adabmDCA.graph import decimate_graph, compute_density
from adabmDCA.statmech import compute_log_likelihood, _update_weights_AIS, compute_entropy, _compute_ess
from adabmDCA.checkpoint import Checkpoint

MAX_EPOCHS = 10000

def fit(
    sampler: Callable,
    chains: torch.Tensor,
    log_weights: torch.Tensor,
    fi_target: torch.Tensor,
    fij_target: torch.Tensor,
    params: Dict[str, torch.Tensor],
    mask: torch.Tensor,
    lr: float,
    nsweeps: int,
    target_pearson: float,
    target_density: float,
    drate: float,
    checkpoint: Checkpoint,
    fi_test: torch.Tensor | None = None,
    fij_test: torch.Tensor | None = None,
    *args, **kwargs,
):
    """Fits an edDCA model on the training data and saves the results in a file.
    
    Args:
        sampler (Callable): Sampling function to be used.
        chains (torch.Tensor): Initialization of the Markov chains.
        log_weights (torch.Tensor): Log-weights of the chains. Used to estimate the log-likelihood.
        fi_target (torch.Tensor): Single-point frequencies of the data.
        fij_target (torch.Tensor): Two-point frequencies of the data.
        params (Dict[str, torch.Tensor]): Initialization of the model's parameters.
        mask (torch.Tensor): Initialization of the coupling matrix's mask.
        lr (float): Learning rate.
        nsweeps (int): Number of Monte Carlo steps to update the state of the model.
        target_pearson (float): Pearson correlation coefficient on the two-points statistics to be reached.
        target_density (float): Target density of the coupling matrix.
        drate (float): Percentage of active couplings to be pruned at each decimation step.
        checkpoint (Checkpoint): Checkpoint class to be used to save the model.
        fi_test (torch.Tensor | None, optional): Single-point frequencies of the test data. Defaults to None.
        fij_test (torch.Tensor | None, optional): Two-point frequencies of the test data. Defaults to None.
        """
    time_start = time.time()
    
    # Check the input sizes
    if fi_target.dim() != 2:
        raise ValueError("fi_target must be a 2D tensor")
    if fij_target.dim() != 4:
        raise ValueError("fij_target must be a 4D tensor")
    if chains.dim() != 3:
        raise ValueError("chains must be a 3D tensor")
    
    L, q = params["bias"].shape
    device = fi_target.device
    dtype = fi_target.dtype
    
    # Get the single-point and two-points frequencies of the simulated data
    pi = get_freq_single_point(data=chains)
    pij = get_freq_two_points(data=chains)
    _, pearson = get_correlation_two_points(fi=fi_target, pi=pi, fij=fij_target, pij=pij)
    if pearson < target_pearson:
        print("Bringing the model to the convergence threshold...")
        chains, params, log_weights, _ = train_graph(
            sampler=sampler,
            chains=chains,
            log_weights=log_weights,
            mask=mask,
            fi=fi_target,
            fij=fij_target,
            fi_test=fi_test,
            fij_test=fij_test,
            params=params,
            nsweeps=nsweeps,
            lr=lr,
            max_epochs=MAX_EPOCHS,
            target_pearson=target_pearson,
            check_slope=False,
            checkpoint=checkpoint,
        )
    
    # Mask for saving only the upper-diagonal matrix
    mask_save = get_mask_save(L, q, device=device)
    
    # Filenames for the decimated parameters and chains
    parent, name = checkpoint.file_paths["params"].parent, checkpoint.file_paths["params"].name
    new_name = name.replace(".dat", "_dec.dat")
    checkpoint.file_paths["params_dec"] = Path(parent).joinpath(new_name)
    
    name = checkpoint.file_paths["chains"].name
    new_name = name.replace(".fasta", "_dec.fasta")
    checkpoint.file_paths["chains_dec"] = Path(parent).joinpath(new_name)
    
    print(f"\nStarting the decimation (target density = {target_density}):")
    with open(checkpoint.file_paths["log"], "a") as f:
        f.write("\nDecimation\n")
        template = "{0:<20} {1:<50}\n"
        f.write(template.format("Target density:", target_density))
        f.write(template.format("Decimation rate:", drate))
        f.write("\n")
        header_string = " ".join([f"{key:<10}" for key in checkpoint.logs.keys()])
        f.write("{0:<10} {1}\n".format("Epoch", header_string))
        
    # Template for wrinting the results
    template = "{0:<15} | {1:<15} | {2:<15} | {3:<15} | {4:<15}"
    density = compute_density(mask)
    count = 0
    checkpoint.checkpt_interval = 10
    
    while density > target_density:
        count += 1
        
        # Store the previous parameters
        prev_params = {key: value.clone() for key, value in params.items()}
        
        # Decimate the model
        params, mask = decimate_graph(
            pij=pij,
            params=params,
            mask=mask,
            drate=drate
        )
        
        # Update the log-weights
        log_weights = _update_weights_AIS(
            prev_params=prev_params,
            curr_params=params,
            chains=chains,
            log_weights=log_weights,
        )
        
        # Equilibrate the model
        chains = sampler(
            chains=chains,
            params=params,
            nsweeps=nsweeps,
        )
        
        # Bring the model at convergence on the graph
        chains, params, log_weights, _ = train_graph(
            sampler=sampler,
            chains=chains,
            log_weights=log_weights,
            mask=mask,
            fi=fi_target,
            fij=fij_target,
            params=params,
            nsweeps=nsweeps,
            lr=lr,
            max_epochs=MAX_EPOCHS,
            target_pearson=target_pearson,
            check_slope=False,
            progress_bar=False,
            checkpoint=None,
        )
        
        # Compute the single-point and two-points frequencies of the simulated data
        pi = get_freq_single_point(data=chains)
        pij = get_freq_two_points(data=chains)
        
        pearson, slope = get_correlation_two_points(fi=fi_target, pi=pi, fij=fij_target, pij=pij)
        density = compute_density(mask)
        logZ = (torch.logsumexp(log_weights, dim=0) - torch.log(torch.tensor(len(chains), device=device, dtype=dtype))).item()
        log_likelihood = compute_log_likelihood(fi=fi_target, fij=fij_target, params=params, logZ=logZ)
        
        print(template.format(f"Step: {count}", f"Density: {density:.3f}", f"LL: {log_likelihood:.3f}", f"Pearson: {pearson:.3f}", f"Slope: {slope:.3f}"))
                
        if checkpoint.check(count, params, chains):
            entropy = compute_entropy(chains=chains, params=params, logZ=logZ)
            ess = _compute_ess(log_weights)
            checkpoint.log(
                {
                    "Epochs": count,
                    "Pearson": pearson,
                    "Slope": slope,
                    "LL_train": log_likelihood,
                    "ESS": ess,
                    "Entropy": entropy,
                    "Density": density,
                    "Time": time.time() - time_start,
                }
            )
            if fi_test is not None and fij_test is not None:
                log_likelihood_test = compute_log_likelihood(fi=fi_test, fij=fij_test, params=params, logZ=logZ)
                checkpoint.log({"LL_test": log_likelihood_test})
            else:
                checkpoint.log({"LL_test": float("nan")})
            
            checkpoint.save(
                params=params,
                mask=torch.logical_and(mask, mask_save),
                chains=chains,
                log_weights=log_weights,
            )
    
    entropy = compute_entropy(chains=chains, params=params, logZ=logZ)
    ess = _compute_ess(log_weights)
    checkpoint.log(
        {
            "Epochs": count,
            "Pearson": pearson,
            "Slope": slope,
            "LL_train": log_likelihood,
            "ESS": ess,
            "Entropy": entropy,
            "Density": density,
            "Time": time.time() - time_start,
        }
    )
    if fi_test is not None and fij_test is not None:
        log_likelihood_test = compute_log_likelihood(fi=fi_test, fij=fij_test, params=params, logZ=logZ)
        checkpoint.log({"LL_test": log_likelihood_test})
    else:
        checkpoint.log({"LL_test": float("nan")})
                
    checkpoint.save(
        params=params,
        mask=torch.logical_and(mask, mask_save),
        chains=chains,
        log_weights=log_weights,
    )
    print(f"Completed, decimated model parameters saved in {checkpoint.file_paths['params_dec']}")


from tqdm.autonotebook import tqdm
import time
from typing import Tuple, Callable, Dict, List
import torch

from adabmDCA.stats import get_freq_single_point, get_freq_two_points, get_correlation_two_points
from adabmDCA.utils import get_mask_save
from adabmDCA.statmech import _update_weights_AIS, compute_log_likelihood, compute_entropy, _compute_ess
from adabmDCA.checkpoint import Checkpoint


@torch.jit.script
def compute_gradient(
    fi: torch.Tensor,
    fij: torch.Tensor,
    pi: torch.Tensor,
    pij: torch.Tensor,
) -> Dict[str, torch.Tensor]:
    """Computes the gradient of the log-likelihood of the model using PyTorch.

    Args:
        fi (torch.Tensor): Single-point frequencies of the data.
        fij (torch.Tensor): Target two-points frequencies.
        pi (torch.Tensor): Single-point marginals of the model.
        pij (torch.Tensor): Two-points marginals of the model.

    Returns:
        Dict[str, torch.Tensor]: Gradient.
    """
    
    grad = {}
    grad["bias"] = fi - pi
    grad["coupling_matrix"] = fij - pij
    
    return grad


def update_params(
    fi: torch.Tensor,
    fij: torch.Tensor,
    pi: torch.Tensor,
    pij: torch.Tensor,
    params: Dict[str, torch.Tensor],
    mask: torch.Tensor,
    lr: float,
) -> Dict[str, torch.Tensor]:
    """Updates the parameters of the model.
    
    Args:
        fi (torch.Tensor): Single-point frequencies of the data.
        fij (torch.Tensor): Two-points frequencies of the data.
        pi (torch.Tensor): Single-point marginals of the model.
        pij (torch.Tensor): Two-points marginals of the model.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        mask (torch.Tensor): Mask of the interaction graph.
        lr (float): Learning rate.
        
    Returns:
        Dict[str, torch.Tensor]: Updated parameters.
    """
    
    # Compute the gradient
    grad = compute_gradient(fi=fi, fij=fij, pi=pi, pij=pij)
    
    # Update parameters
    with torch.no_grad():
        for key in params:
            params[key] += lr * grad[key]
        params["coupling_matrix"] *= mask # Remove autocorrelations
    
    return params


def train_graph(
    sampler: Callable,
    chains: torch.Tensor,
    mask: torch.Tensor,
    fi: torch.Tensor,
    fij: torch.Tensor,
    params: Dict[str, torch.Tensor],
    nsweeps: int,
    lr: float,
    max_epochs: int,
    target_pearson: float,
    fi_test: torch.Tensor | None = None,
    fij_test: torch.Tensor | None = None,
    checkpoint: Checkpoint | None = None,
    check_slope: bool = False,
    log_weights: torch.Tensor | None = None,
    progress_bar: bool = True,
) -> Tuple[torch.Tensor, Dict[str, torch.Tensor], torch.Tensor, Dict[str, List[float]]]:
    """Trains the model on a given graph until the target Pearson correlation is reached or the maximum number of epochs is exceeded.

    Args:
        sampler (Callable): Sampling function.
        chains (torch.Tensor): Markov chains simulated with the model.
        mask (torch.Tensor): Mask encoding the sparse graph.
        fi (torch.Tensor): Single-point frequencies of the data.
        fij (torch.Tensor): Two-point frequencies of the data.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        nsweeps (int): Number of Gibbs steps for each gradient estimation.
        lr (float): Learning rate.
        max_epochs (int): Maximum number of gradient updates to be done.
        target_pearson (float): Target Pearson coefficient.
        fi_test (torch.Tensor | None, optional): Single-point frequencies of the test data. Defaults to None.
        fij_test (torch.Tensor | None, optional): Two-point frequencies of the test data. Defaults to None.
        checkpoint (Checkpoint | None, optional): Checkpoint class to be used for saving the model. Defaults to None.
        check_slope (bool, optional): Whether to take into account the slope for the convergence criterion or not. Defaults to False.
        log_weights (torch.Tensor, optional): Log-weights used for the online computation of the log-likelihood. Defaults to None.
        progress_bar (bool, optional): Whether to display a progress bar or not. Defaults to True.

    Returns:
        Tuple[torch.Tensor, Dict[str, torch.Tensor], torch.Tensor, Dict[str, List[float]]]: Updated chains and parameters, log-weights for the log-likelihood computation.
    """
    device = fi.device
    dtype = fi.dtype
    L, q = fi.shape
    time_start = time.time()
    
    # log_weights used for the online computing of the log-likelihood
    if log_weights is None:
        log_weights = torch.zeros(len(chains), device=device, dtype=dtype)
    logZ = (torch.logsumexp(log_weights, dim=0) - torch.log(torch.tensor(len(chains), device=device, dtype=dtype))).item()
    log_likelihood = compute_log_likelihood(fi=fi, fij=fij, params=params, logZ=logZ)
    
    # Compute the single-point and two-points frequencies of the simulated data
    pi = get_freq_single_point(data=chains)
    pij = get_freq_two_points(data=chains)
    history = {
        "epochs": [],
        "pearson": [],
        "slope": [],
    }
    
    def halt_condition(epochs, pearson, slope, check_slope):
        c1 = pearson < target_pearson
        c2 = epochs < max_epochs
        if check_slope:
            c3 = abs(slope - 1.) > 0.1
        else:
            c3 = False
        return not c2 * ((not c1) * c3 + c1)
    
    # Mask for saving only the upper-diagonal coupling matrix
    mask_save = get_mask_save(L, q, device=device)

    pearson, slope = get_correlation_two_points(fij=fij, pij=pij, fi=fi, pi=pi)
    epochs = 0
    if progress_bar: 
        pbar = tqdm(
            initial=max(0, float(pearson)),
            total=target_pearson,
            colour="red",
            dynamic_ncols=True,
            leave=False,
            ascii="-#",
            bar_format="{desc} {percentage:.2f}%[{bar}] Pearson: {n:.3f}/{total_fmt} [{elapsed}]"
        )
        pbar.set_description(f"Epochs: {epochs} - LL: {log_likelihood:.2f}")
    
    while not halt_condition(epochs, pearson, slope, check_slope):
        
        # Store the previous parameters
        params_prev = {key: value.clone() for key, value in params.items()}
        
        # Update the parameters
        params = update_params(
            fi=fi,
            fij=fij,
            pi=pi,
            pij=pij,
            params=params,
            mask=mask,
            lr=lr,
        )
        
        # Compute the weights for the AIS
        log_weights = _update_weights_AIS(
            prev_params=params_prev,
            curr_params=params,
            chains=chains,
            log_weights=log_weights,
        )
        
        # Update the Markov chains
        chains = sampler(chains=chains, params=params, nsweeps=nsweeps)
        epochs += 1
        
        # Compute the single-point and two-points frequencies of the simulated data
        pi = get_freq_single_point(data=chains)
        pij = get_freq_two_points(data=chains)
        pearson, slope = get_correlation_two_points(fij=fij, pij=pij, fi=fi, pi=pi)
        logZ = (torch.logsumexp(log_weights, dim=0) - torch.log(torch.tensor(len(chains), device=device, dtype=dtype))).item()
        log_likelihood = compute_log_likelihood(fi=fi, fij=fij, params=params, logZ=logZ)
        
        if progress_bar:
            pbar.n = min(max(0, float(pearson)), target_pearson)
            pbar.set_description(f"Epochs: {epochs} - LL: {log_likelihood:.2f}")
        
        history["epochs"].append(epochs)
        history["pearson"].append(pearson)
        history["slope"].append(slope)
            
        if checkpoint is not None:
            entropy = compute_entropy(chains=chains, params=params, logZ=logZ)
            ess = _compute_ess(log_weights)
            if fi_test is not None and fij_test is not None:
                log_likelihood_test = compute_log_likelihood(fi=fi_test, fij=fij_test, params=params, logZ=logZ)
            else:
                log_likelihood_test = float("nan")
            checkpoint.log(
                {
                    "Epochs": epochs,
                    "Pearson": pearson,
                    "Slope": slope,
                    "LL_train": log_likelihood,
                    "LL_test": log_likelihood_test,
                    "ESS": ess,
                    "Entropy": entropy,
                    "Density": 1.0,
                    "Time": time.time() - time_start,
                }
            )
            
            # Save the model if a checkpoint is reached
            if checkpoint.check(epochs, params, chains):
                checkpoint.save(
                    params=params,
                    mask=mask_save,
                    chains=chains,
                    log_weights=log_weights,
                )
                
    if progress_bar:
        pbar.close()
    
    if checkpoint is not None:            
        checkpoint.save(
            params=params,
            mask=mask_save,
            chains=chains,
            log_weights=log_weights,
        )
        
    return chains, params, log_weights, history



import argparse
from pathlib import Path
import matplotlib.pyplot as plt

from adabmDCA.fasta import get_tokens
from adabmDCA.io import load_params
from adabmDCA.utils import get_device, get_dtype
from adabmDCA.parser import add_args_contacts
from adabmDCA.dca import get_contact_map
from adabmDCA.plot import plot_contact_map

# import command-line input arguments
def create_parser():
    # Important arguments
    parser = argparse.ArgumentParser(description='Computes the Frobenius norm matrix extracted from a DCA model.')
    parser = add_args_contacts(parser)
    
    return parser


def main():    
    # Parse arguments
    parser = create_parser()
    args = parser.parse_args()
    
    print("\n" + "".join(["*"] * 10) + f" Computing the Frobenius norm " + "".join(["*"] * 10) + "\n")
    
    # Set the device
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)
    
    # Check if the parameters file exists
    if not Path(args.path_params).exists():
        raise FileNotFoundError(f"Parameters file {args.path_params} not found.")
    
    # Import parameters
    tokens = get_tokens(args.alphabet)
    print(f"Loading parameters from {args.path_params}...")
    params = load_params(args.path_params, tokens=tokens, device=device, dtype=dtype)
    Fapc = get_contact_map(params, tokens)
    
    # Save the results
    print("Saving results...")
    if args.label is not None:
        fname_out = args.output / Path(f"{args.label}_contact_map.txt")
    else:
        fname_out = args.output / Path(f"contact_map.txt")
      
    with open(fname_out, "w") as f:  
        for i in range(Fapc.shape[0]):
            for j in range(Fapc.shape[1]):
                f.write(f"{i},{j},{Fapc[i, j]}\n")
                
    # plot the contact map into a file
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(dpi=150, figsize=(6, 5))
    ax = fig.add_subplot(111)
    ax = plot_contact_map(ax, Fapc)
    fig.tight_layout()
    fig.savefig(fname_out.with_suffix(".png"))

    print(f"Process completed. Results saved in {fname_out}")
    
if __name__ == "__main__":
    main()     
    


from pathlib import Path
import numpy as np
from typing import Tuple

import torch

# Default alphabets
TOKENS_PROTEIN = "-ACDEFGHIKLMNPQRSTVWY"
TOKENS_RNA = "-ACGU"
TOKENS_DNA = "-ACGT"


def get_tokens(alphabet: str) -> str:
    """Converts the alphabet into the corresponding tokens.

    Args:
        alphabet (str): Alphabet to be used for the encoding. It can be either "protein", "rna", "dna" or a custom string of tokens.

    Returns:
        str: Tokens of the alphabet.
    """
    assert isinstance(alphabet, str), "Argument 'alphabet' must be of type str"
    if alphabet == "protein":
        return TOKENS_PROTEIN
    elif alphabet == "rna":
        return TOKENS_RNA
    elif alphabet == "dna":
        return TOKENS_DNA
    else:
        return alphabet
    
    
def encode_sequence(sequence: str | np.ndarray | list, tokens: str) -> np.ndarray:
    """Encodes a sequence or a list of sequences into a numeric format.

    Args:
        sequence (str | np.ndarray | list): Input sequence.
        tokens (str): Alphabet to be used for the encoding.

    Returns:
        np.ndarray: Encoded sequence or sequences.
    """
    letter_map = {l : n for n, l in enumerate(tokens)}
    
    def _encode(sequece):
        return [letter_map[l] for l in sequece]
    
    if isinstance(sequence, str):
        return np.array(_encode(sequence))
    elif isinstance(sequence, np.ndarray):
        sequence = list(sequence)
        return np.array(list(map(_encode, sequence)))
    elif isinstance(sequence, list):
        return np.array(list(map(_encode, sequence)))
    else:        
        raise ValueError("Input sequence must be either a string or a numpy array.")


def decode_sequence(sequence: list | np.ndarray | torch.Tensor, tokens: str) -> str | np.ndarray:
    """Takes a numeric sequence or list of seqences in input an returns the corresponding string encoding.

    Args:
        sequence (np.ndarray): Input sequences. Can be either a 1D or a 2D iterable.
        tokens (str): Alphabet to be used for the encoding.

    Returns:
        str | np.ndarray: string or array of strings with the decoded input.
    """
    if isinstance(sequence, list):
        sequence = np.array(sequence)
    elif isinstance(sequence, torch.Tensor):
        sequence = sequence.cpu().numpy()
    assert isinstance(sequence, np.ndarray), "Input sequence must be either a numpy array, a list or a torch tensor."
    sequence = sequence.astype(int)
    
    def _decode(sequence):
        return ''.join([tokens[aa] for aa in sequence])
    
    if sequence.ndim == 1:
        return _decode(sequence)
    elif sequence.ndim == 2:
        sequence = list(sequence)
        return np.array(list(map(_decode, sequence)))
    else:
        raise ValueError("Input sequence must be either a 1D or a 2D array.")


def import_from_fasta(
    fasta_name: str | Path,
    tokens: str | None = None,
    filter_sequences: bool = False,
    remove_duplicates: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """Import sequences from a fasta file. The following operations are performed:
    - If 'tokens' is provided, encodes the sequences in numeric format.
    - If 'filter_sequences' is True, removes the sequences whose tokens are not present in the alphabet.
    - If 'remove_duplicates' is True, removes the duplicated sequences.

    Args:
        fasta_name (str | Path): Path to the fasta file.
        tokens (str | None, optional): Alphabet to be used for the encoding. If provided, encodes the sequences in numeric format.
        filter_sequences (bool, optional): If True, removes the sequences whose tokens are not present in the alphabet. Defaults to False.
        remove_duplicates (bool, optional): If True, removes the duplicated sequences. Defaults to True.

    Raises:
        RuntimeError: The file is not in fasta format.

    Returns:
        Tuple[np.ndarray, np.ndarray]: headers, sequences.
    """
    # Import headers and sequences
    sequences = []
    names = []
    seq = ''
    with open(fasta_name, 'r') as f:
        first_line = f.readline()
        if not first_line.startswith('>'):
            raise RuntimeError(f"The file {fasta_name} is not in a fasta format.")
        f.seek(0)
        for line in f:
            if not line.strip():
                continue
            if line.startswith('>'):
                if seq:
                    sequences.append(seq)
                header = line[1:].strip()
                names.append(header)
                seq = ''
            else:
                seq += line.strip()
    if seq:
        sequences.append(seq)
    
    # Filter sequences
    if filter_sequences:
        if tokens is None:
            raise ValueError("Argument 'tokens' must be provided if 'filter_sequences' is True.")
        tokens = get_tokens(tokens)
        tokens_list = [a for a in tokens]
        clean_names = []
        clean_sequences = []
        for n, s in zip(names, sequences):
            good_sequence = np.full(shape=(len(s),), fill_value=False)
            splitline = np.array([a for a in s])
            for token in tokens_list:
                good_sequence += (token == splitline)
            if np.all(good_sequence):
                if n == "":
                    n = "unknown_sequence"
                clean_names.append(n)
                clean_sequences.append(s)
            else:
                print(f"Unknown token found: removing sequence {n}")
        names = np.array(clean_names)
        sequences = np.array(clean_sequences)
        
    else:
        names = np.array(names)
        sequences = np.array(sequences)
    
    # Remove duplicates
    if remove_duplicates:
        sequences, unique_ids = np.unique(sequences, return_index=True)
        names = names[unique_ids]
    
    if (tokens is not None) and (len(sequences) > 0):
        sequences = encode_sequence(sequences, tokens)
    
    return names, sequences
    
    
def write_fasta(
    fname: str,
    headers: np.ndarray | list,
    sequences: np.ndarray,
    numeric_input: bool = False,
    remove_gaps: bool = False,
    tokens: str = "protein",
):
    """Generate a fasta file with the input sequences.

    Args:
        fname (str): Name of the output fasta file.
        headers (np.ndarray | list): Array or list of sequences' headers.
        sequences (np.ndarray): Array of sequences.
        numeric_input (bool, optional): Whether the sequences are in numeric (encoded) format or not. Defaults to False.
        remove_gaps (bool, optional): If True, removes the gap from the alignment. Defaults to False.
        tokens (str): Alphabet to be used for the encoding. Defaults to protein.
    """
    tokens = get_tokens(tokens)

    if numeric_input:
        # Decode the sequences
        seqs_decoded = decode_sequence(sequences, tokens)
    else:
        seqs_decoded = sequences.copy()
    if remove_gaps:
        seqs_decoded = np.vectorize(lambda s: s.replace("-", ""))(seqs_decoded)
        
    with open(fname, 'w') as f:
        for name_seq, seq in zip(headers, seqs_decoded):
            f.write('>' + name_seq + '\n')
            f.write(seq)
            f.write('\n')


def compute_weights(
    data: np.ndarray | torch.Tensor,
    th: float = 0.8,
    device: torch.device = torch.device("cpu"),
    dtype: torch.dtype = torch.float32,
) -> torch.Tensor:
    """Computes the weight to be assigned to each sequence 's' in 'data' as 1 / n_clust, where 'n_clust' is the number of sequences
    that have a sequence identity with 's' >= th.

    Args:
        data (np.ndarray | torch.Tensor): Encoded input dataset.
        th (float, optional): Sequence identity threshold for the clustering. Defaults to 0.8.
        device (toch.device, optional): Device. Defaults to "cpu".
        dtype (torch.dtype, optional): Data type. Defaults to torch.float32.

    Returns:
        torch.Tensor: Array with the weights of the sequences.
    """
    if isinstance(data, torch.Tensor):
        data = data.to(device)
    else:
        data = torch.tensor(data, device=device)

    assert len(data.shape) == 2, "'data' must be a 2-dimensional array"
    _, L = data.shape

    @torch.jit.script
    def get_sequence_weight(s: torch.Tensor, data: torch.Tensor, L: int, th: float):
        seq_id = torch.sum(s == data, dim=1) / L
        n_clust = torch.sum(seq_id > th)
        
        return 1.0 / n_clust

    weights = torch.vstack([get_sequence_weight(s, data, L, th) for s in data])
    
    return weights.to(dtype)


def validate_alphabet(sequences: np.ndarray, tokens: str):
    """Check if the chosen alphabet is compatible with the input sequences.

    Args:
        sequences (np.ndarray): Input sequences.
        tokens (str): Alphabet to be used for the encoding.

    Raises:
        KeyError: The chosen alphabet is incompatible with the Multi-Sequence Alignment.
    """
    all_char = "".join(sequences)
    tokens_data = "".join(sorted(set(all_char)))
    sorted_tokens = "".join(sorted(tokens))
    if not sorted_tokens == tokens_data:
        raise KeyError(f"The chosen alphabet is incompatible with the Multi-Sequence Alignment. The missing tokens are: {[c for c in tokens_data if c not in sorted_tokens]}. Current alphabet: {tokens}")


import argparse
from pathlib import Path

import torch

from adabmDCA.fasta import get_tokens
from adabmDCA.utils import get_device, get_dtype
from adabmDCA.io import load_params, import_from_fasta
from adabmDCA.fasta import encode_sequence, decode_sequence
from adabmDCA.functional import one_hot
from adabmDCA.statmech import compute_energy
from adabmDCA.parser import add_args_dms


# import command-line input arguments
def create_parser():
    parser = argparse.ArgumentParser(description="Generates the Deep Mutational Scanning of a given wild type.")
    parser = add_args_dms(parser)
    
    return parser


def main():    
    # Parse arguments
    parser = create_parser()
    args = parser.parse_args()
    
    print("\n" + "".join(["*"] * 10) + f" Generating DMS " + "".join(["*"] * 10) + "\n")
    
    # Set the device
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)
    
    # Check if the data file exists
    if not Path(args.data).exists():
        raise FileNotFoundError(f"Data file {args.data} not found.")
    
    # Check if the parameters file exists
    if not Path(args.path_params).exists():
        raise FileNotFoundError(f"Parameters file {args.path_params} not found.")
    
    # import data and parameters
    tokens = get_tokens(args.alphabet)
    names, sequences = import_from_fasta(args.data)
    wt_name = names[0]
    # remove non-alphanumeric characters from wt name
    wt_name = "".join(e for e in wt_name if e.isalnum())
    wt_seq = torch.tensor(encode_sequence(sequences[0], tokens))
    
    print(f"Loading parameters from {args.path_params}...")
    params = load_params(args.path_params, tokens=tokens, device=device, dtype=dtype)
    L, q = params["bias"].shape
    
    # generate DMS
    dms = []
    site_list = []
    old_residues = []
    new_residues = []
    
    print("Generating the single mutant library...")
    for i in range(L):
        for a in range(q):
            if wt_seq[i] != a:
                seq = wt_seq.clone()
                seq[i] = a
                dms.append(seq)
                site_list.append(i)
                old_residues.append(tokens[wt_seq[i]])
                new_residues.append(tokens[a])
    
    print("Computing the DCA scores...")
    dms = torch.vstack(dms).to(device=device)
    dms = one_hot(dms, num_classes=q).to(dtype)
    energies = compute_energy(dms, params)
    energy_wt = compute_energy(one_hot(wt_seq.view(1, -1).to(device), num_classes=q).to(dtype), params)
    deltaE = energies - energy_wt
    
    dms = torch.argmax(dms, -1).cpu().numpy()
    dms_decoded = decode_sequence(dms, tokens)
    
    print("Saving the results...")
    folder = args.output
    folder.mkdir(parents=True, exist_ok=True)
    fname_out = folder / Path(f"{wt_name}_DMS.fasta")
    
    with open(fname_out, "w") as f:
        for i, res_old, res_new, e, seq in zip(site_list, old_residues, new_residues, deltaE, dms_decoded):
            f.write(f">{res_old}{i}{res_new} | DCAscore: {e:.3f}\n")
            f.write(seq + "\n")
    
    print(f"Process completed. Results saved in {fname_out}")
    

if __name__ == "__main__":
    main()


from typing import Callable, Dict

import torch

from adabmDCA.training import train_graph
from adabmDCA.checkpoint import Checkpoint

    
def fit(
    sampler: Callable,
    fi_target: torch.Tensor,
    fij_target: torch.Tensor,
    params: Dict[str, torch.Tensor],
    mask: torch.Tensor,
    chains: torch.Tensor,
    log_weights: torch.Tensor,
    target_pearson: float,
    nsweeps: int,
    nepochs: int,
    lr: float,
    fi_test: torch.Tensor | None = None,
    fij_test: torch.Tensor | None = None,
    checkpoint: Checkpoint | None = None,
    *args, **kwargs,
) -> None:
    """Trains a bmDCA model on the input MSA and saves the results in a file.

    Args:
        sampler (Callable): Sampling function to be used.
        fi_target (torch.Tensor): Single-point frequencies of the data.
        fij_target (torch.Tensor): Two-point frequencies of the data.
        params (Dict[str, torch.Tensor]): Initialization of the model's parameters.
        mask (torch.Tensor): Initialization of the coupling matrix's mask.
        chains (torch.Tensor): Initialization of the Markov chains.
        log_weights (torch.Tensor): Log-weights of the chains. Used to estimate the log-likelihood.
        target_pearson (float): Pearson correlation coefficient on the two-points statistics to be reached.
        nsweeps (int): Number of Monte Carlo steps to update the state of the model.
        nepochs (int): Maximum number of epochs to be performed.
        lr (float): Learning rate.
        fi_test (torch.Tensor | None, optional): Single-point frequencies of the test data. Defaults to None.
        fij_test (torch.Tensor | None, optional): Two-point frequencies of the test data. Defaults to None.
        checkpoint (Checkpoint | None, optional): Checkpoint class to be used to save the model. Defaults to None.
    """
    
    # Check the input sizes
    if fi_target.dim() != 2:
        raise ValueError("fi_target must be a 2D tensor")
    if fij_target.dim() != 4:
        raise ValueError("fij_target must be a 4D tensor")
    if chains.dim() != 3:
        raise ValueError("chains must be a 3D tensor")
    
    # Training loop    
    train_graph(
        sampler=sampler,
        chains=chains,
        log_weights=log_weights,
        mask=mask,
        fi=fi_target,
        fij=fij_target,
        fi_test=fi_test,
        fij_test=fij_test,
        params=params,
        nsweeps=nsweeps,
        lr=lr,
        max_epochs=nepochs,
        target_pearson=target_pearson,
        check_slope=False,
        checkpoint=checkpoint,
    )



from adabmDCA.fasta import (
    import_from_fasta,
    get_tokens,
    write_fasta,
    encode_sequence,
    decode_sequence,
)
from adabmDCA.stats import (
    get_freq_single_point,
    get_freq_two_points,
    get_freq_three_points,
    get_correlation_two_points,
)
from adabmDCA.io import (
    load_params,
    save_params,
)
from adabmDCA.sampling import (
    get_sampler,
    gibbs_sampling,
    metropolis,
)
from adabmDCA.functional import one_hot

__all__ = [
    "import_from_fasta",
    "get_tokens",
    "write_fasta",
    "encode_sequence",
    "decode_sequence",
    "get_freq_single_point",
    "get_freq_two_points",
    "get_freq_three_points",
    "get_correlation_two_points",
    "load_params",
    "save_params",
    "get_sampler",
    "gibbs_sampling",
    "metropolis",
    "one_hot",
]


from typing import Dict, Tuple
import torch


@torch.jit.script
def compute_Dkl_activation(
    fij: torch.Tensor,
    pij: torch.Tensor,
) -> torch.Tensor:
    """Computes the Kullback-Leibler divergence matrix of all the possible couplings.
    
    Args:
        fij (torch.Tensor): Two-point frequences of the dataset.
        pij (torch.Tensor): Two-point marginals of the model.
    
    Returns:
        torch.Tensor: Kullback-Leibler divergence matrix.
    """
    L = fij.shape[0]
    # Compute the Dkl of each coupling
    Dkl = fij * (torch.log(fij) - torch.log(pij)) + (1. - fij) * (torch.log(1. - fij) - torch.log(1. - pij))
    # The auto-correlations have not to be considered
    Dkl[torch.arange(L), :, torch.arange(L), :] = -float("inf")
    
    return Dkl


def update_mask_activation(
    Dkl: torch.Tensor,
    mask: torch.Tensor,
    nactivate: int,
) -> torch.Tensor:
    """Updates the mask by removing the nactivate couplings with the smallest Dkl.
    
    Args:
        Dkl (torch.Tensor): Kullback-Leibler divergence matrix.
        mask (torch.Tensor): Mask.
        nactivate (int): Number of couplings to be activated at each graph update.
    
    Returns:
        torch.Tensor: Updated mask.
    """
    # Flatten the Dkl tensor and sort it in descending order
    Dkl_flat_sorted, _ = torch.sort(Dkl.flatten(), descending=True)
    # Get the threshold value for the top nactivate*2 elements (since Dkl is symmetric)
    Dkl_th = Dkl_flat_sorted[2 * nactivate]
    # Update the mask where Dkl is greater than the threshold
    mask = torch.where(Dkl > Dkl_th, torch.tensor(1, device=Dkl.device, dtype=Dkl.dtype), mask)
    
    return mask


@torch.jit.script
def activate_graph(
    mask: torch.Tensor,
    fij: torch.Tensor,
    pij: torch.Tensor,
    nactivate: int,
) -> torch.Tensor: 
    """Updates the interaction graph by activating a maximum of nactivate couplings.

    Args:
        mask (torch.Tensor): Mask.
        fij (torch.Tensor): Two-point frequencies of the dataset.
        pij (torch.Tensor): Two-point marginals of the model.
        nactivate (int): Number of couplings to activate.
        
    Returns:
        torch.Tensor: Updated mask.
    """
    
    # Compute the Kullback-Leibler divergence of all the couplings
    Dkl = compute_Dkl_activation(fij=fij, pij=pij)
    # Update the graph
    mask = update_mask_activation(Dkl=Dkl, mask=mask, nactivate=nactivate)
    
    return mask


@torch.jit.script
def compute_sym_Dkl(
    params: Dict[str, torch.Tensor],
    pij: torch.Tensor,
) -> torch.Tensor:
    """Computes the symmetric Kullback-Leibler divergence matrix between the initial distribution and the same 
    distribution once removing one coupling J_ij(a, b).

    Args:
        params (Dict[ste, torch.Tensor]): Parameters of the model.
        pij (torch.Tensor): Two-point marginal probability distribution.

    Returns:
        torch.Tensor: Kullback-Leibler divergence matrix.
    """
    
    exp_J = torch.exp(-params["coupling_matrix"])
    Dkl = pij * params["coupling_matrix"] * (1. - exp_J / (pij * (exp_J - 1.) + 1.))
    
    return Dkl


@torch.jit.script
def compute_Dkl_decimation(
    params: Dict[str, torch.Tensor],
    pij: torch.Tensor,
) -> torch.Tensor:
    """Computes the Kullback-Leibler divergence matrix between the initial distribution and the same 
    distribution once removing one coupling J_ij(a, b).

    Args:
        params (Dict[ste, torch.Tensor]): Parameters of the model.
        pij (torch.Tensor): Two-point marginal probability distribution.

    Returns:
        torch.Tensor: Kullback-Leibler divergence matrix.
    """
    
    exp_J = torch.exp(-params["coupling_matrix"])
    Dkl = pij * params["coupling_matrix"] + torch.log(exp_J * pij + 1 - pij)
    
    return Dkl


def update_mask_decimation(
    mask: torch.Tensor,
    Dkl: torch.Tensor,
    drate: float,
) -> torch.Tensor:
    """Updates the mask by removing the n_remove couplings with the smallest Dkl.

    Args:
        mask (torch.Tensor): Mask.
        Dkl (torch.Tensor): Kullback-Leibler divergence matrix.
        drate (float): Percentage of active couplings to be pruned at each decimation step.

    Returns:
        torch.Tensor: Updated mask.
    """
    
    n_remove = int((mask.sum().item() // 2) * drate) * 2
    # Only consider the active couplings
    Dkl_active = torch.where(mask, Dkl, float("inf")).view(-1)
    _, idx_remove = torch.topk(-Dkl_active, n_remove)
    mask = mask.view(-1).scatter_(0, idx_remove, 0).view(mask.shape)
    
    return mask


def decimate_graph(
    pij: torch.Tensor,
    params: Dict[str, torch.Tensor],
    mask: torch.Tensor,
    drate: float,
) -> Tuple[Dict[str, torch.Tensor], torch.Tensor]:
    """Performs one decimation step and updates the parameters and mask.

    Args:
        pij (torch.Tensor): Two-point marginal probability distribution.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        mask (torch.Tensor): Mask.
        drate (float): Percentage of active couplings to be pruned at each decimation step.

    Returns:
        Tuple[Dict[str, torch.Tensor], torch.Tensor]: Updated parameters and mask.
    """
    
    Dkl = compute_Dkl_decimation(params=params, pij=pij)
    mask = update_mask_decimation(mask=mask, Dkl=Dkl, drate=drate)
    params["coupling_matrix"] *= mask
    
    return params, mask


@torch.jit.script
def compute_density(mask: torch.Tensor) -> float:
    """Computes the density of active couplings in the coupling matrix.

    Args:
        mask (torch.Tensor): Mask.

    Returns:
        float: Density.
    """
    L, q, _, _ = mask.shape
    density = mask.sum() / (q**2 * L * (L-1))
    
    return density.item()


from setuptools import setup, find_packages

setup(
    name='adabmDCA',
    version='0.5.0',
    author='Lorenzo Rosset, Roberto Netti, Anna Paola Muntoni, Francesco Zamponi, Martin Weigt',
    maintainer='Lorenzo Rosset',
    author_email='rosset.lorenzo@gmail.com',
    description='Python implementation of Direct Coupling Analysis',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/spqb/adabmDCA',
    packages=find_packages(include=['adabmDCA', 'adabmDCA.*']),
    include_package_data=True,
    python_requires='>=3.10',
    license_files=["LICENSE"],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'adabmDCA=adabmDCA.cli:main',
        ],
    },
    install_requires=[
        'matplotlib>=3.8.0',
        'numpy>=1.26.4',
        'pandas>=2.2.2',
        'torch>=2.1.0',
        'tqdm>=4.66.6',
        'h5py>=3.12.1',
        'wandb>=0.12.0',
    ],
)



from typing import Dict, Any
from abc import ABC, abstractmethod
import torch
import h5py
import wandb

from adabmDCA.io import save_chains, save_params
from adabmDCA.statmech import _get_acceptance_rate


class Checkpoint(ABC):
    """Helper class to save the model's parameters and chains at regular intervals during training and to log the
    progress of the training.
    """
    def __init__(
        self,
        file_paths: dict,
        tokens: str,
        args: dict,
        params: Dict[str, torch.Tensor] | None = None,
        chains: torch.Tensor | None = None,
        use_wandb: bool = False,
    ):
        """Initializes the Checkpoint class.

        Args:
            file_paths (dict): Dictionary containing the paths of the files to be saved.
            tokens (str): Alphabet to be used for encoding the sequences.
            args (dict): Dictionary containing the arguments of the training.
            params (Dict[str, torch.Tensor] | None, optional): Parameters of the model. Defaults to None.
            chains (torch.Tensor | None, optional): Chains. Defaults to None.
            use_wandb (bool, optional): Whether to use Weights & Biases for logging. Defaults to False.
        """
        if not isinstance(args, dict):
            args = vars(args)
            
        self.file_paths = file_paths
        self.tokens = tokens
        
        self.wandb = use_wandb
        if self.wandb:
            wandb.init(project="adabmDCA", config=args)
            
        if params is not None:
            self.params = {key: value.clone() for key, value in params.items()}
        else:
            self.params = None
        if chains is not None:
            self.chains = chains.clone()
        else:
            self.chains = None
        self.max_epochs = args["nepochs"]
        self.checkpt_interval = 50
        
        self.logs = {
            "Epochs": 0,
            "Pearson": 0.0,
            "Slope": 0.0,
            "LL_train": 0.0,
            "LL_test": 0.0,
            "ESS": 0.0,
            "Entropy": 0.0,
            "Density": 0.0,
            "Time": 0.0,
        }
        
        template = "{0:<20} {1:<50}\n"  
        with open(file_paths["log"], "w") as f:
            if args["label"] is not None:
                f.write(template.format("label:", args["label"]))
            else:
                f.write(template.format("label:", "N/A"))
            
            f.write(template.format("model:", str(args["model"])))
            f.write(template.format("input MSA:", str(args["data"])))
            f.write(template.format("alphabet:", args["alphabet"]))
            f.write(template.format("sampler:", args["sampler"]))
            f.write(template.format("nchains:", args["nchains"]))
            f.write(template.format("nsweeps:", args["nsweeps"]))
            f.write(template.format("lr:", args["lr"]))
            f.write(template.format("pseudo count:", args["pseudocount"]))
            f.write(template.format("data type:", args["dtype"]))
            f.write(template.format("target Pearson Cij:", args["target"]))
            if args["model"] == "eaDCA":
                f.write(template.format("gsteps:", args["gsteps"]))
                f.write(template.format("factivate:", args["factivate"]))
            f.write(template.format("random seed:", args["seed"]))
            f.write("\n")
            # write the header of the log file
            header_string = " ".join([f"{key:<10}" for key in self.logs.keys()])
            f.write(header_string + "\n")
        
        
    def log(
        self,
        record: Dict[str, Any],
    ) -> None:
        """Adds a key-value pair to the log dictionary

        Args:
            record (Dict[str, Any]): Key-value pairs to be added to the log dictionary.
        """
        for key, value in record.items():
            if key not in self.logs.keys():
                raise ValueError(f"Key {key} not recognized.")
        
            if isinstance(value, torch.Tensor):
                self.logs[key] = value.item()
            else:
                self.logs[key] = value
                
        if self.wandb:
            wandb.log(self.logs)        
        out_string = " ".join([f"{value:<10.3f}" if isinstance(value, float) else f"{value:<10}" for value in self.logs.values()])
        with open(self.file_paths["log"], "a") as f:
            f.write(out_string + "\n")
    
    @abstractmethod
    def check(
        self,
        updates: int,
        *args,
        **kwargs,
    ) -> bool:
        """Checks if a checkpoint has been reached.
        
        Args:
            updates (int): Number of gradient updates performed.

        Returns:
            bool: Whether a checkpoint has been reached.
        """
        pass
        
        
    @abstractmethod 
    def save(
        self,
        params: Dict[str, torch.Tensor],
        mask: torch.Tensor,
        chains: torch.Tensor,
        log_weights: torch.Tensor,
    ) -> None:
        """Saves the chains and the parameters of the model.

        Args:
            params (Dict[str, torch.Tensor]): Parameters of the model.
            mask (torch.Tensor): Mask of the model's coupling matrix representing the interaction graph
            chains (torch.Tensor): Chains.
            log_weights (torch.Tensor): Log of the chain weights. Used for AIS.
        """
        pass
    
    
class LinearCheckpoint(Checkpoint):
    def __init__(
        self,
        file_paths: dict,
        tokens: str,
        args: dict,
        params: Dict[str, torch.Tensor] | None = None,
        chains: torch.Tensor | None = None,
        checkpt_interval: int = 50,
        use_wandb: bool = False,
        **kwargs,
    ):
        super().__init__(
            file_paths=file_paths,
            tokens=tokens,
            args=args,
            params=params,
            chains=chains,
            use_wandb=use_wandb,
        )
        self.checkpt_interval = checkpt_interval
        
    
    def check(
        self,
        updates: int,
        *args,
        **kwargs,
    ) -> bool:
        """Checks if a checkpoint has been reached.
        
        Args:
            updates (int): Number of gradient updates performed.

        Returns:
            bool: Whether a checkpoint has been reached.
        """
        return (updates % self.checkpt_interval == 0) or (updates == self.max_epochs)
    
    
    def save(
        self,
        params: Dict[str, torch.Tensor],
        mask: torch.Tensor,
        chains: torch.Tensor,
        log_weights: torch.Tensor,
    ) -> None:
        """Saves the chains and the parameters of the model.

        Args:
            params (Dict[str, torch.Tensor]): Parameters of the model.
            mask (torch.Tensor): Mask of the model's coupling matrix representing the interaction graph
            chains (torch.Tensor): Chains.
            log_weights (torch.Tensor): Log of the chain weights. Used for AIS.
        """            
        save_params(fname=self.file_paths["params"], params=params, mask=mask, tokens=self.tokens)
        save_chains(fname=self.file_paths["chains"], chains=chains.argmax(dim=-1), tokens=self.tokens, log_weights=log_weights)
            
            
class AcceptanceCheckpoint(Checkpoint):
    def __init__(
        self,
        file_paths: Dict,
        tokens: str,
        args: Dict,
        params: Dict[str, torch.Tensor],
        chains: torch.Tensor | None = None,
        target_acc_rate: float = 0.5,
        use_wandb: bool = False,
        **kwargs,
    ):
        super().__init__(
            file_paths=file_paths,
            tokens=tokens,
            args=args,
            params=params,
            chains=chains,
            use_wandb=use_wandb,
        )
        self.target_acc_rate = target_acc_rate
        self.num_saved_models = 0
        # Create a .h5 archive for storing the history of the parameters
        self.file_paths["params_history"] = self.file_paths["params"].with_suffix(".h5")
        with h5py.File(self.file_paths["params_history"], "w") as f:
            f["alphabet"] = self.tokens
            f.create_group("update_{0}".format(self.logs["Epochs"]))
            for key, value in params.items():
                f["update_{0}".format(self.logs["Epochs"])].create_dataset(key, data=value.cpu().numpy())
        
    def check(
        self,
        updates: int,
        curr_params: Dict[str, torch.Tensor],
        curr_chains: torch.Tensor,
        *args,
        **kwargs,
    ) -> bool:
        """Checks if a checkpoint has been reached by computing the acceptance rate of swapping the 
        configurations of the present model and the last saved model.
        
        Args:
            updates (int): Number of gradient updates performed.
            curr_params (Dict[str, torch.Tensor]): Current parameters of the model.
            curr_chains (torch.Tensor): Current chains of the model.

        Returns:
            bool: Whether a checkpoint has been reached.
        """
        acc_rate = _get_acceptance_rate(
            prev_params=self.params,
            curr_params=curr_params,
            prev_chains=self.chains,
            curr_chains=curr_chains,
        )
        return (acc_rate < self.target_acc_rate) or (updates == self.max_epochs)
    
    
    def save(
        self,
        params: Dict[str, torch.Tensor],
        mask: torch.Tensor,
        chains: torch.Tensor,
        log_weights: torch.Tensor,
        *args,
        **kwargs,
    ) -> None:
        """Saves the chains and the parameters of the model and appends the current parameters to the
        file containing the parameters history.

        Args:
            params (Dict[str, torch.Tensor]): Parameters of the model.
            mask (torch.Tensor): Mask of the model's coupling matrix representing the interaction graph.
            chains (torch.Tensor): Chains.
            log_weights (torch.Tensor): Log of the chain weights. Used for AIS.
        """
            
        # Store the current parameters and chains
        self.params = {key: value.clone() for key, value in params.items()}
        self.chains = chains.clone()
        # Append the current parameters to the history
        with h5py.File(self.file_paths["params_history"], "a") as f:
            f.create_group("update_{0}".format(self.logs["Epochs"]))
            for key, value in params.items():
                f["update_{0}".format(self.logs["Epochs"])].create_dataset(key, data=value.cpu().numpy())
        # Save the current parameters and chains
        save_params(fname=self.file_paths["params"], params=params, mask=mask, tokens=self.tokens)
        save_chains(fname=self.file_paths["chains"], chains=chains.argmax(dim=-1), tokens=self.tokens, log_weights=log_weights)
            
            
def get_checkpoint(chpt: str) -> Checkpoint:
    if chpt == "linear":
        return LinearCheckpoint
    elif chpt == "acceptance":
        return AcceptanceCheckpoint
    else:
        raise ValueError(f"Checkpoint type {chpt} not recognized.")        


import argparse
import logging
import os
import torch
from adabmDCA.fasta import import_from_fasta, get_tokens, write_fasta
from adabmDCA.cobalt import run_cobalt
from adabmDCA.parser import add_args_profmark

# import command-line input arguments
def create_parser():
    parser = argparse.ArgumentParser(description="Splits a multi-sequence alignment FASTA file into training and test sets using the Cobalt algorithm described in Petti et al (2022).")
    parser = add_args_profmark(parser)
    
    return parser

def main(args):
    print("\n" + "".join(["*"] * 10) + f" Splitting dataset into train and test sets " + "".join(["*"] * 10) + "\n")
    logging.basicConfig(level=logging.INFO)
    
    # Set the device
    if args.device == "cuda" and not torch.cuda.is_available():
        args.device = "cpu"
    device = torch.device(args.device)
    logging.info(f"Using device: {device}")
    
    # check if the output directory exists
    if os.path.dirname(args.output_prefix):
        if not os.path.exists(os.path.dirname(args.output_prefix)):
            os.makedirs(os.path.dirname(args.output_prefix))
    # check if the input MSA exists
    if not os.path.exists(args.input_msa):
        raise FileNotFoundError(f"Input MSA file {args.input_msa} does not exist.")
    
    # load the MSA
    logging.info(f"Loading MSA from {args.input_msa}")
    tokens = get_tokens(args.alphabet)
    headers, msa = import_from_fasta(args.input_msa, tokens)
    msa = torch.tensor(msa, device=device)
    logging.info(f"Loaded {len(headers)} sequences of length {len(msa[0])} from {args.input_msa}")
    
    # run the Cobalt algorithm
    geom_mean = 0
    rnd_gen = torch.Generator(device=device).manual_seed(args.seed)
    for i in range(args.bestof):
        logging.info(f"Running Cobalt algorithm iteration {i+1}/{args.bestof}")
        headers_train_prop, train_prop, headers_test_prop, test_prop = run_cobalt(
            headers,
            msa,
            args.t1,
            args.t2,
            args.t3,
            args.maxtrain,
            args.maxtest,
            rnd_gen,
        )
        geom_mean_prop = len(train_prop) * len(test_prop)
        logging.info(f"---> Iteration {i+1}: len(train)={len(train_prop)}, len(test)={len(test_prop)}")
        if geom_mean_prop > geom_mean:
            geom_mean = geom_mean_prop
            headers_train, train, headers_test, test = (
                headers_train_prop,
                train_prop,
                headers_test_prop,
                test_prop,
            )
    if geom_mean == 0:
        logging.warning("FAILED: No split was found for t1={}, t2={}, t3={}".format(args.t1, args.t2, args.t3))
        return
    logging.info(f"Best iteration: len(train)={len(train)}, len(test)={len(test)}, geom_mean={geom_mean}")
    # write the training and test sets to files
    write_fasta(
        os.path.join(args.output_prefix + ".train.fasta"),
        headers_train,
        train.cpu().numpy(),
        tokens,
    )
    write_fasta(
        os.path.join(args.output_prefix + ".test.fasta"),
        headers_test,
        test.cpu().numpy(),
        tokens,
    )
    logging.info(f"Training and test sets written in {os.path.dirname(args.output_prefix)}")
    
if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    main(args)


import numpy as np
import torch
from typing import Tuple, Dict
from adabmDCA.stats import get_covariance_matrix


def get_seqid(
    s1: torch.Tensor,
    s2: torch.Tensor | None = None,
    average: bool = False,
) -> torch.Tensor | Tuple[torch.Tensor, torch.Tensor]:
    """
    When average is True:
    - If s2 is provided, computes the mean and the standard deviation of the mean sequence identity between two sets of one-hot encoded sequences.
    - If s2 is a single sequence (L, q), it computes the mean and the standard deviation of the mean sequence identity between the dataset s1 and s2.
    - If s2 is none, computes the mean and the standard deviation of the mean of the sequence identity between s1 and a permutation of s1.
    
    When average is False it returns the array of sequence identities.

    Args:
        s1 (torch.Tensor): Sequence dataset 1.
        s2 (torch.Tensor | None): Sequence dataset 2. Defaults to None.
        average (bool): Whether to return the average and standard deviation of the sequence identity or the array of sequence identities.

    Returns:
        torch.Tensor | Tuple[torch.Tensor, torch.Tensor]: List of sequence identities or mean sequence identity and standard deviation of the mean.
    """
    if len(s1.shape) == 2:
        s1 = s1.unsqueeze(0)
    if s2 is None:
        s2 = s1[torch.randperm(s1.shape[0])]
    if len(s2.shape) == 2:
        s2 = s2.unsqueeze(0)
        
    s1 = s1.view(s1.shape[0], -1)
    s2 = s2.view(s2.shape[0], -1)
    
    seqids = (s1 * s2).sum(1)
    if average is False:
        return seqids
    else:
        mean_seqid = seqids.mean()
        if len(seqids) == 1:
            std_seqid = torch.tensor(0.0, device=seqids.device)
        else:
            std_seqid = seqids.std() / np.sqrt(len(seqids))

        return mean_seqid, std_seqid


def set_zerosum_gauge(params: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    """Sets the zero-sum gauge on the coupling matrix.
    
    Args:
        params (Dict[str, torch.Tensor]): Parameters of the model.
        
    Returns:
        Dict[str, torch.Tensor]: Parameters with fixed gauge.
    """
    coupling_matrix = params["coupling_matrix"]
    coupling_matrix -= coupling_matrix.mean(dim=1, keepdim=True) + \
                       coupling_matrix.mean(dim=3, keepdim=True) - \
                       coupling_matrix.mean(dim=(1, 3), keepdim=True)
    
    params["coupling_matrix"] = coupling_matrix
    
    return params


def get_contact_map(
    params: Dict[str, torch.Tensor],
    tokens: str,
) -> np.ndarray:
    """
    Computes the contact map from the model coupling matrix.

    Args:
        params (Dict[str, torch.Tensor]): Model parameters.
        tokens (str): Alphabet.

    Returns:
        np.ndarray: Contact map.
    """
    q = params["coupling_matrix"].shape[1]

    # Zero-sum gauge  
    params = set_zerosum_gauge(params)
    
    # Get index of the gap symbol
    gap_idx = tokens.index("-")
    
    Jij = params["coupling_matrix"]
    # Take all the entries of the coupling matrix except where the gap is involved
    mask = torch.arange(q) != gap_idx
    Jij_reduced = Jij[:, mask, :, :][:, :, :, mask]

    # Compute the Frobenius norm
    cm = torch.sqrt(torch.square(Jij_reduced).sum([1, 3]))
    # Set to zero the diagonal
    cm = cm - torch.diag(cm.diag())
    # Compute the average-product corrected Frobenius norm
    Fapc = cm - torch.outer(cm.sum(1), cm.sum(0)) / cm.sum()
    # set to zero the diagonal
    Fapc = Fapc - torch.diag(Fapc.diag())

    return Fapc.cpu().numpy()


def get_mf_contact_map(
    data: torch.Tensor,
    tokens: str,
    weights: torch.Tensor | None = None
) -> np.ndarray:
    """
    Computes the contact map from the model coupling matrix.

    Args:
        data (torch.Tensor): Input one-hot data tensor.
        tokens (str): Alphabet.
        weights (torch.Tensor | None): Weights for the data points. Defaults to None.

    Returns:
        np.ndarray: Contact map.
    """
    device = data.device
    dtype = data.dtype
    L, q = data.shape[1], data.shape[2]
    # Get index of the gap symbol
    gap_idx = tokens.index("-")
    
    # Compute the covariance matrix
    Cij = get_covariance_matrix(data, weights=weights)
    shrink = 4.5 / torch.sqrt(torch.tensor(data.shape[0], dtype=dtype, device=device)) * torch.eye(Cij.shape[0], device=device, dtype=dtype)
    Cij += shrink
        
    # Invert the covariance matrix to get the coupling matrix
    Jij = -torch.linalg.inv(Cij)

    # partial correlation coefficient
    Jij_diag = torch.diag(Jij)
    pcc = Jij / torch.sqrt(Jij_diag[:, None] * Jij_diag[None, :])
    pcc = pcc.reshape(L, q, L, q)
    
    # Take all the entries of the coupling matrix except where the gap is involved
    mask = torch.arange(q) != gap_idx
    pcc = pcc[:, mask, :, :][:, :, :, mask]
    
    # Compute the Frobenius norm
    F = torch.sqrt(torch.square(pcc).sum([1, 3]))
    # Set to zero the diagonal
    F = F - torch.diag(F.diag())
    # Compute the average-product corrected Frobenius norm
    Fapc = F - (F.sum(1, keepdim=True) * F.sum(0, keepdim=True)) / F.sum()
    # Set to zero the diagonal
    Fapc = Fapc - torch.diag(Fapc.diag())

    return Fapc.cpu().numpy()


import argparse
from pathlib import Path

def add_args_dca(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    dca_args = parser.add_argument_group("General DCA arguments")
    dca_args.add_argument("-d", "--data",         type=Path,  required=True,        help="Filename of the dataset to be used for training the model.")
    dca_args.add_argument("-o", "--output",       type=Path,  default='DCA_model',  help="(Defaults to DCA_model). Path to the folder where to save the model.")
    dca_args.add_argument("-m", "--model",        type=str,   default="bmDCA",      help="(Defaults to bmDCA). Type of model to be trained.", choices=["bmDCA", "eaDCA", "edDCA"])
    # Optional arguments
    dca_args.add_argument("-t", "--test",         type=Path,  default=None,         help="(Defaults to None). Filename of the dataset to be used for testing the model.")
    dca_args.add_argument("-p", "--path_params",  type=Path,  default=None,         help="(Defaults to None) Path to the file containing the model's parameters. Required for restoring the training.")
    dca_args.add_argument("-c", "--path_chains",  type=Path,  default=None,         help="(Defaults to None) Path to the fasta file containing the model's chains. Required for restoring the training.")
    dca_args.add_argument("-l", "--label",        type=str,   default=None,         help="(Defaults to None). If provoded, adds a label to the output files inside the output folder.")
    dca_args.add_argument("--alphabet",           type=str,   default="protein",    help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    dca_args.add_argument("--lr",                 type=float, default=0.01,         help="(Defaults to 0.01). Learning rate.")
    dca_args.add_argument("--nsweeps",            type=int,   default=10,           help="(Defaults to 10). Number of sweeps for each gradient estimation.")
    dca_args.add_argument("--sampler",            type=str,   default="gibbs",      help="(Defaults to gibbs). Sampling method to be used.", choices=["metropolis", "gibbs"])
    dca_args.add_argument("--nchains",            type=int,   default=10000,        help="(Defaults to 10000). Number of Markov chains to run in parallel.")
    dca_args.add_argument("--target",             type=float, default=0.95,         help="(Defaults to 0.95). Pearson correlation coefficient on the two-sites statistics to be reached.")
    dca_args.add_argument("--nepochs",            type=int,   default=50000,        help="(Defaults to 50000). Maximum number of epochs allowed.")
    dca_args.add_argument("--pseudocount",        type=float, default=None,         help="(Defaults to None). Pseudo count for the single and two-sites statistics. Acts as a regularization. If None, it is set to 1/Meff.")
    dca_args.add_argument("--seed",               type=int,   default=0,            help="(Defaults to 0). Seed for the random number generator.")
    dca_args.add_argument("--wandb",              action="store_true",              help="If provided, logs the training on Weights and Biases.")
    dca_args.add_argument("--device",             type=str,   default="cuda",       help="(Defaults to cuda). Device to be used.")
    dca_args.add_argument("--dtype",              type=str,   default="float32",    help="(Defaults to float32). Data type to be used.")
    
    return parser


def add_args_reweighting(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    weight_args = parser.add_argument_group("Sequence reweighting arguments")
    weight_args.add_argument("-w", "--weights",      type=Path,  default=None,         help="(Defaults to None). Path to the file containing the weights of the sequences. If None, the weights are computed automatically.")
    weight_args.add_argument("--clustering_seqid",   type=float, default=0.8,          help="(Defaults to 0.8). Sequence Identity threshold for clustering. Used only if 'weights' is not provided.")
    weight_args.add_argument("--no_reweighting",     action="store_true",              help="If provided, the reweighting of the sequences is not performed.")

    return parser


def add_args_checkpoint(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    checkpoint_args = parser.add_argument_group("Checkpoint arguments")
    checkpoint_args.add_argument("--checkpoints",     type=str,   default="linear",     help="(Defaults to 'linear'). Choses the type of checkpoint criterium to be used.", choices=["linear", "acceptance"])
    checkpoint_args.add_argument("--target_acc_rate", type=float, default=0.5,          help="(Defaults to 0.5). Target acceptance rate for deciding when to save the model when the 'acceptance' checkpoint is used.")
    
    return parser


def add_args_eaDCA(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    eadca_args = parser.add_argument_group("eaDCA arguments")
    eadca_args.add_argument("--gsteps",           type=int,   default=10,           help="(Defaults to 10). Number of gradient updates to be performed on a given graph.")
    eadca_args.add_argument("--factivate",        type=float, default=0.001,        help="(Defaults to 0.001). Fraction of inactive couplings to be proposed for activation at each graph update.")

    return parser


def add_args_edDCA(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    eddca_args = parser.add_argument_group("edDCA arguments")
    eddca_args.add_argument("--density",          type=float, default=0.02,         help="(Defaults to 0.02). Target density to be reached.")
    eddca_args.add_argument("--drate",            type=float, default=0.01,         help="(Defaults to 0.01). Fraction of remaining couplings to be pruned at each decimation step.")

    return parser


def add_args_train(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser = add_args_dca(parser)
    parser = add_args_eaDCA(parser)
    parser = add_args_edDCA(parser)
    parser = add_args_reweighting(parser)
    parser = add_args_checkpoint(parser)
    
    return parser


def add_args_energies(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument("-d", "--data",         type=Path,   required=True,      help="Path to the fasta file containing the sequences.")
    parser.add_argument("-p", "--path_params",  type=Path,   required=True,      help="Path to the file containing the parameters of DCA model.")
    parser.add_argument("-o", "--output",       type=Path,   required=True,      help="Path to the folder where to save the output.")
    parser.add_argument("--alphabet",           type=str,    default="protein",  help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    parser.add_argument("--device",             type=str,    default="cuda",     help="(Defaults to cuda). Device to be used.")
    parser.add_argument("--dtype",              type=str,    default="float32",  help="(Defaults to float32). Data type to be used.")

    return parser


def add_args_contacts(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument("-p", "--path_params",  type=Path,   required=True,          help="Path to the file containing the parameters of DCA model to sample from.")
    parser.add_argument("-o", "--output",       type=Path,   required=True,          help="Path to the folder where to save the output.")
    parser.add_argument("-l", "--label",        type=str,    default=None,           help="(Defaults to None). If provoded, adds a label to the output files inside the output folder.")
    parser.add_argument("--alphabet",           type=str,    default="protein",      help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    parser.add_argument("--device",             type=str,    default="cuda",         help="(Defaults to cuda). Device to be used.")
    parser.add_argument("--dtype",              type=str,    default="float32",      help="(Defaults to float32). Data type to be used.")
    
    return parser


def add_args_dms(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument("-d", "--data",         type=Path,   required=True,      
                        help="Path to the fasta file containing wild type sequence. If more than one sequence is present, the first one is used.")
    parser.add_argument("-p", "--path_params",  type=Path,   required=True,      help="Path to the file containing the parameters of DCA model to sample from.")
    parser.add_argument("-o", "--output",       type=Path,   required=True,      help="Path to the folder where to save the output.")
    parser.add_argument("--alphabet",           type=str,    default="protein",  help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    parser.add_argument("--device",             type=str,    default="cuda",     help="(Defaults to cuda). Device to be used.")
    parser.add_argument("--dtype",              type=str,    default="float32",  help="(Defaults to float32). Data type to be used.")

    return parser


def add_args_sample(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument("-p", "--path_params",  type=Path,   required=True,      help="Path to the file containing the parameters of DCA model to sample from.")
    parser.add_argument("-d", "--data",         type=Path,   required=True,      help="Path to the file containing the data to sample from.")
    parser.add_argument("-o", "--output",       type=Path,   required=True,      help="Path to the folder where to save the output.")
    parser.add_argument("--ngen",               type=int,    required=True,      help="Number of sequences to be generated.") 
    
    # Optional arguments
    parser.add_argument("-l", "--label",        type=str,    default="sampling", help="(Defaults to 'sampling'). Label to be used for the output files.")
    parser.add_argument("--nmeasure",           type=int,    default=10000,      help="(Defaults to min(10000, len(data)). Number of data sequences to use for computing the mixing time.")
    parser.add_argument("--nmix",               type=int,    default=2,          help="(Defaults to 2). Number of mixing times used to generate 'ngen' sequences starting from random.")
    parser.add_argument("--max_nsweeps",        type=int,    default=1000,       help="(Defaults to 1000). Maximum number of chain updates.")
    parser.add_argument("--alphabet",           type=str,    default="protein",  help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    parser.add_argument("--sampler",            type=str,    default="gibbs",    help="(Defaults to gibbs). Sampling method to be used. Choose between 'metropolis' and 'gibbs'.")
    parser.add_argument("--beta",               type=float,  default=1.0,        help="(Defaults to 1.0). Inverse temperature for the sampling.")
    parser.add_argument("--pseudocount",        type=float,  default=None,       help="(Defaults to None). Pseudocount for the single and two points statistics used during the training. If None, 1/Meff is used.")
    parser.add_argument("--device",             type=str,    default="cuda",     help="(Defaults to cuda). Device to perform computations on.")
    parser.add_argument("--dtype",              type=str,    default="float32",  help="(Defaults to float32). Data type to be used.")
    
    parser = add_args_reweighting(parser)
    
    return parser

def add_args_tdint(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument("-p", "--path_params",      type=Path,   required=True,           help="Path to the file containing the parameters of DCA model to sample from.")
    parser.add_argument("-d", "--data",             type=Path,   required=True,           help="Path to the file containing the data to sample from.")
    parser.add_argument("-t", "--path_targetseq",   type=Path,   required=True,           help="Path to the file containing the target sequence.")
    parser.add_argument("-o", "--output",           type=Path,   default='DCA_model',     help="Path to the folder where to save the output.")
   
    # Optional arguments
    parser.add_argument("-c", "--path_chains",  type=Path,   default=None,            help="(Defaults to None). Path to the fasta file containing the model's chains.")
    parser.add_argument("-l", "--label",        type=str,    default="entropy",       help="(Defaults to 'entropy'). Label to be used for the output files.")
    parser.add_argument("--nchains",            type=int,    default=10000,           help="(Defaults to 10000). Number of chains to be used.") 
    parser.add_argument("--theta_max",          type=float,  default=5,               help="(Defaults to 5). Maximum integration strength") 
    parser.add_argument("--nsteps",             type=int,    default=100,             help="(Defaults to 100). Number of integration steps.")
    parser.add_argument("--nsweeps",            type=int,    default=100,             help="(Defaults to 100). Number of chain updates for each integration step.")
    parser.add_argument("--nsweeps_theta",      type=int,    default=100,             help="(Defaults to 100). Number of chain updates to equilibrate chains at theta_max.")
    parser.add_argument("--nsweeps_zero",       type=int,    default=100,             help="(Defaults to 100). Number of chain updates to equilibrate chains at theta=0.")
    parser.add_argument("--alphabet",           type=str,    default="protein",       help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    parser.add_argument("--nepochs",            type=int,    default=50000,           help="(Defaults to 50000). Maximum number of epochs allowed.")
    parser.add_argument("--sampler",            type=str,    default="gibbs",         help="(Defaults to gibbs). Sampling method to be used. Choose between 'metropolis' and 'gibbs'.")
    parser.add_argument("--seed",               type=int,    default=0,               help="(Defaults to 0). Seed for the random number generator.")
    parser.add_argument("--device",             type=str,    default="cuda",          help="(Defaults to cuda). Device to perform computations on.")
    parser.add_argument("--dtype",              type=str,    default="float32",       help="(Defaults to float32). Data type to be used.")
    
    return parser

def add_args_reintegration(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument("--reint",  type=Path,   required=True,  help="Path to the fasta file containing the reintegrated sequences.")
    parser.add_argument("--adj",    type=Path,   required=True,  help="Path to the file containing the adjustment vector.")
    parser.add_argument("--lambda_", type=float,    required=True,  help="Lambda parameter for the reintegration.")

    return parser

def add_args_profmark(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument("output_prefix", type=str, help="Prefix for the output files.")
    parser.add_argument("input_msa", type=str, help="Fasta file containing the multiple sequence alignment.")
    parser.add_argument("-t1", type=float, default=0.5, help="No sequence in S (the candidate training set) has more than this fraction of its residues identical to any sequence in T (the candidate test set). Default is 0.5.")
    parser.add_argument("-t2", type=float, default=0.5, help="No pair of test sequences has more than this value fractional identity. Default is 0.5.")
    parser.add_argument("-t3", type=float, default=1.0, help="No pair of training sequences has more than this value fractional identity. Default is 1.0.")
    parser.add_argument("--bestof", type=int, default=1, help="Runs the algorithm n times and returns the one that maximizes |S| * |T|. Default is 1.")
    parser.add_argument("--maxtrain", type=int, default=None, help="Maximum number of sequences in the training set.")
    parser.add_argument("--maxtest", type=int, default=None, help="Maximum number of sequences in the test set.")
    parser.add_argument("--alphabet", type=str, default="protein", help="Alphabet to use for encoding the sequences. Choose among 'protein', 'rna', 'dna' or a user-defined alphabet. Default is 'protein'.")
    parser.add_argument("--seed", type=int, default=42, help="Random seed for reproducibility. Default is 42.")
    parser.add_argument("--device", type=str, default="cuda", help="Device to use for computation. Default is 'cuda'.")

    return parser